/************************************************************************
RWhois Software 

Copyright (c) 1994 Scott Williamson and Mark Kosters

Users may distribute the binary and source code to third parties provided
that the copyright notice and this statement appears on all copies and 
that no charge is made for such copies.  

This software is provided AS IS with no warranties of any kind.  The authors
shall have no liability with respect to the infringement of copyrights,
trade secrets or any patents by this file or any part thereof.  In no
event will the authors be liable for any lost revenue or profits or
other special, indirect and consequential damages.

Comments and additions should be sent to the authors:

                    Scott Williamson
                    Network Solutions
                    505 Huntmar Park Drive
                    Herndon, VA 22071
                    (703) 742-4820
                    scottw@internic.net

                    Mark Kosters
                    Network Solutions
                    505 Huntmar Park Drive
                    Herndon, VA 22071
                    (703) 742-4795
                    markk@internic.net


Developed By: 
  Mark Kosters
***************************************************************************/
#include "objects.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif /* HAVE_STRINGS_H */

#include "attributes.h"
#include "client_msgs.h"
#include "commands.h"
#include "defines.h"
#include "display_conf.h"
#include "log.h"
#include "misc.h"
#include "strutil.h"


/* local static variables  */
static object_struct *object_head = (object_struct *) NULL;
static llist_struct  *object_by_type[NUM_OBJECTS];

/* -------------------- LOCAL FUNCTIONS ------------------- */

/* sets the object type */
static object_type
get_object_type(type)
  char *type;
{
  trim(type);

  if (STR_EQ(type, "OBJECT"))
  {
    return OBJECT;
  }
  else if (STR_EQ(type, "REFERRAL"))
  {
    return REFERRAL;
  }
  else if (STR_EQ(type, "SOA"))
  {
    return SOA;
  }

  log_error("Unknown template type %s in object definition file", type);

  return NUM_OBJECTS;
}

/* read_command_from_object: Given a command tagged line from a
     schema/object file, parse the line and add the command to the
     global command list.  The command line has the format
     "command_name:command_length" */
static void
read_command_from_object(object, str)
  object_struct *object;
  char          *str;
{
  char                  command[MAX_TEMPLATE_DESC];
  char                  length_str[MAX_TEMPLATE_DESC];
  int                   length;

  if (parse_line(str, command, length_str))
  {
    length = atoi(length_str);
  }
  else
  {
    strncpy(command, str, MAX_TEMPLATE_DESC);
    length = strlen(command);
  }

  append_command(command, length, NULL, NULL, TRUE, COMMAND_OBJECT, object,
                 NULL, NULL);
}

/* -------------------- PUBLIC FUNCTIONS ------------------ */

/* reads the main schema/object configuration file.  Returns TRUE if
     successful, FALSE otherwise.  */
int
read_objects(file)
  char  *file;
{
  char          line[MAX_LINE];
  char          tag[MAX_TEMPLATE_DESC];
  char          datum[MAX_TEMPLATE_DESC];
  FILE          *fp     = NULL;
  object_struct *cur    = NULL;
  object_struct *new    = NULL;

  if ((fp = fopen(file, "r")) == NULL)
  {
    log_error("could not open schema file '%s': %s", file,
              l_strerror(errno));
    return FALSE;
  }

  /* make new template record */
  if ((new = NEW_OBJECT) == (object_struct *)NULL)
  {
    log_error("read_objects: memory allocation problem");
    return(FALSE);
  }

  /* if object_head *isn't* NULL, then we have a problem */
  if (object_head == NULL)
  {
    object_head = new;
    cur         = new;
  }
  else
  {
    log_error("read_object() called without clearing the object list first");
    return FALSE;
  }
  
  while ((fgets(line, MAX_LINE, fp)) != NULL)
  {

    /* store last record; prepare for reading the next record */
    if (new_record(line))
    {

      /* read the attributes for this object */
      if (!read_attributes(cur))
      {
        /* if the attribute read failed, log an error and then discard
           the object */
        log_error("could not load attributes for '%s'", cur->name);
        bzero(cur, sizeof(cur));
        continue;
      }

      /* add the object to the "by type" lists */
      if (cur->type < NUM_OBJECTS)
      {
        object_by_type[cur->type] = append_llist(object_by_type[cur->type],
                                                 cur);
        /* this is redundant, but what the hey */
        object_by_type[EVERY_OBJECT]
          = append_llist(object_by_type[EVERY_OBJECT], cur);
      }
      

      /* make new template record */
      if ((new = NEW_OBJECT) == NULL)
      {
        log_error("read_objects: memory_allocation_problem");
        return FALSE;
      }
      
      cur->nxt = new;
      cur      = new;

      continue;
    }       

    if (parse_line(line, tag, datum))
    {
      if (STR_EQ(tag, O_NAME))
      {
        cur->name = NEW_STRING(datum);
      }
      else if (STR_EQ(tag, O_ATTRIBUTEDEF))
      {
        cur->attribute_fname = NEW_STRING(datum);
      }
      else if (STR_EQ(tag, O_LOCALDB))
      {
        cur->local_db = NEW_STRING(datum);
      }
      else if (STR_EQ(tag, O_CACHEDB))
      {
        cur->cache_db = NEW_STRING(datum);
      }
      else if (STR_EQ(tag, O_DBDIR))
      {
        cur->dbdir = NEW_STRING(datum);
      }
      else if (STR_EQ(tag, O_DISPLAY))
      {
        cur->output_format
          = read_display_format_from_object(cur->output_format,
                                            datum);
      }
      else if (STR_EQ(tag, O_DESCRIPTION))
      {
        cur->description = NEW_STRING(datum);
      }
      else if (STR_EQ(tag, O_COMMAND))
      {
        read_command_from_object(cur, datum);
      }
      else if (STR_EQ(tag, O_TYPE))
      {
        cur->type = get_object_type(datum);
      }
    }
  }

  /* commit last entry */
  if (read_attributes(cur))
  {
    /* add to the "by_type" lists */
    object_by_type[cur->type] = append_llist(object_by_type[cur->type], cur);
    object_by_type[EVERY_OBJECT]
      = append_llist(object_by_type[EVERY_OBJECT], cur);
  } 

  fclose (fp);

  return TRUE;
}

/* return_objects_by_type: returns a llist of objects specified by
     type */
llist_struct *
return_objects_by_type(type)
  object_type type; 
{
  return(object_by_type[type]);
}

/* returns all the objects defined */
llist_struct *
return_all_defined_objects(str)
  char *str;
{
  object_struct *object;

  if (str && *str && ((object = return_object_by_command(str)) != NULL))
  {
    /* return a linked list of one element, object */
    return(append_llist(NULL, object));
  }
  /* if a blank or nonexistant string is passed in, all object match */
  else if (!str || !*str)
  {
    return (return_objects_by_type(EVERY_OBJECT));
  }

  /* tried to find the object by the command, and failed */
  return NULL;
}


/* returns attributes of object by name */
attribute_struct *
get_object_attributes_by_object_name(object_name)
  char *object_name;
{
  object_struct *cur;

  cur = object_head;
  
  while (cur != NULL)
  {
    if (STR_EQ(cur->name, object_name))
    {
      return(cur->attributes);
    }
    cur = cur-> nxt;
  }
  return (NULL);
}

/* returns object by name */
object_struct *
get_object_by_object_name(object_name)
  char *object_name;
{
  object_struct *cur;

  cur = object_head;
  
  while (cur != NULL)
  {
    if (STR_EQ(cur->name, object_name))
    {
      return (cur);
    }

    cur = cur-> nxt;
  }

  return (NULL);
}


/* builds linked list or returns a pointer to linked list. Returns
     templates that match criterion, NULL if none exist */
llist_struct *
setup_search_objects(query_item, type)
  query_item_struct *query_item;
  object_type       type;
{
  query_item_struct *cur;

  if (type == REFERRAL || type == SOA)
    return (return_objects_by_type(type));
  
  /* first pass - check to see if any object pointers defined  - return it */
  cur = query_item;
  while (cur != NULL)
  {
    if (cur->object != NULL)
    {
      return (append_llist (NULL, cur->object));
    }

    cur = cur->nxt;
  }

  /* otherwise, we give up and return all normal objects */
  return(return_objects_by_type(OBJECT));
}

/* returns the object based on the first field (should be Object-type:
   [object]), NULL if world is on the right place to be */
object_struct *
get_object_from_record (fields)
  field_struct *fields;
{
  attribute_struct *attrib;

  /* quick core dump fix */
  if (!fields) return (NULL);
  
  if ((attrib = get_attribute_info_from_internal_id(fields->id)) != NULL)
  {
    return (return_object_by_command(attrib->object_name));
  }
  return (NULL);
}

/* returns the pointer to the attributes of that designated object
   based on first tag.  Returns ptr if exists, NULL if not */
attribute_struct *
get_object_attributes_by_tag(tag, datum)
  char *tag;
  char *datum;

{
  llist_struct      *object_list    = NULL;
  object_struct     *object         = NULL;
  attribute_struct  *attrib         = NULL;
  int               found = FALSE;

  object_list = return_all_defined_objects(datum);
  
  while (object_list !=  NULL && !found)
  {
    object = object_list->datum;
    attrib = object->attributes;

    while (attrib != NULL && !found)
    {
      if (STR_EQ(tag, attrib->name)  ||
          (STR_EQ(tag, attrib->name_alias) && attrib->def_object))
      {
        found = TRUE;
      }
      else
      {
        attrib=attrib->nxt;
      }
    }

    object_list = object_list->nxt;
  }
  if (found)
  {
    return (object->attributes);
  }
  else
  {
    return (NULL);
  }
}


